from mlc_tools.base.model import Model
from mlc_tools.core.class_ import Class
from mlc_tools.core.function import Function
from mlc_tools.core.object import Object, Objects, AccessSpecifier


RETAIN_INT = '++this->_reference_counter;'
RETAIN_ATOMIC = '_reference_counter.fetch_add(1, std::memory_order_relaxed);'
RELEASE_INT = '''
--this->_reference_counter;
auto counter = this->_reference_counter;
if(counter == 0)
{
    delete this;
}
return counter;
'''
RELEASE_ATOMIC = '''
auto ret = _reference_counter.fetch_sub(1, std::memory_order_acq_rel);
if(ret == 1)
{
    delete this;
}
return ret;
'''


class GeneratorRefCounter(object):
    def __init__(self):
        self.prefer_type = 'int'
        self.retain = RETAIN_INT
        self.release = RELEASE_INT

    def generate(self, model: Model):
        if not model.generate_ref_counter:
            return

        if model.side == 'server':
            self.prefer_type = 'std::atomic<int>'
            self.retain = RETAIN_ATOMIC
            self.release = RELEASE_ATOMIC

        for cls in model.classes:
            if not cls.superclasses and cls.type != 'enum' and not cls.is_abstract:
                self._add(cls)

    def _add(self, cls: Class):
        if not cls.has_member_with_name('_reference_counter'):
            ref_counter = Object()
            ref_counter.name = '_reference_counter'
            ref_counter.type = self.prefer_type
            ref_counter.initial_value = '1'
            ref_counter.is_runtime = True
            ref_counter.access = AccessSpecifier.private
            cls.members.append(ref_counter)
        if not cls.has_method_with_name('retain'):
            retain = Function()
            retain.name = 'retain'
            retain.return_type = Objects.VOID
            retain.operations.extend(self.retain.split('\n'))
            cls.functions.append(retain)
        if not cls.has_method_with_name('release'):
            release = Function()
            release.name = 'release'
            release.return_type = Objects.INT
            release.operations.extend(self.release.split('\n'))
            cls.functions.append(release)



from ..base.generator_operator_equals import GeneratorOperatorEqualsBase
from ..core.class_ import Class
from ..core.function import Function
from ..core.object import Objects


class GeneratorOperatorEquals(GeneratorOperatorEqualsBase):

    def __init__(self):
        GeneratorOperatorEqualsBase.__init__(self)

    def get_equal_method_name(self):
        return 'operator =='

    def get_not_equal_method_name(self):
        return 'operator !='

    def get_compare_method_pattern(self, cls, member):
        body_line = 'result = result && this->{0} == rhs.{0};'
        body_line_ptr = \
            'result = result && ((this->{0} == rhs.{0}) || ' \
            '(this->{0} != nullptr && rhs.{0} != nullptr && *this->{0} == *rhs.{0}));'
        return body_line_ptr if member.is_pointer or member.is_link else body_line

    def get_not_equal_method_operation(self):
        return 'return !(*this == rhs);'

    def add_copy_constructor(self, cls: Class):
        copy_constructor = Function()
        copy_constructor.name = cls.name
        copy_constructor.return_type = Objects.VOID
        copy_constructor.args.append(['rhs', GeneratorOperatorEqualsBase.get_const_ref(cls)])
        # operator.operations.append(self.get_not_equal_method_operation())

        copy_constructor.operations.append(f'this->operator=(rhs);')
        cls.functions.append(copy_constructor)

    def add_move_constructor(self, cls):
        pass

    def add_copy_operator(self, cls):
        copy_operator = Function()
        copy_operator.name = 'operator ='
        copy_operator.return_type = GeneratorOperatorEqualsBase.get_const_ref(cls)
        copy_operator.args.append(['rhs', GeneratorOperatorEqualsBase.get_const_ref(cls)])

        if cls.superclasses and cls.superclasses[0] != 'SerializedObject':
            parent = f'this->{cls.superclasses[0]}::operator=(rhs);'
            copy_operator.operations.append(parent)

        for member in cls.members:
            if member.is_static:
                continue
            if self.model.side == 'server' and member.name == '_reference_counter':
                copy_operator.operations.append(f'this->{member.name}.store(rhs.{member.name}.load());')
            else:
                copy_operator.operations.append(f'this->{member.name} = rhs.{member.name};')
        copy_operator.operations.append('return *this;')
        cls.functions.append(copy_operator)
